/*
LICENSE
-------
The Anti-Viral License

You may use this source code in any way you like provided that this source code
retains its original licensing terms.  Derivative works of this  source code
must also be licensed under this license,  however,  this  license does not
forbid compiling this source code alongside  source  code  of other licenses
provided that those licenses do not require a change in the licensing of this
source code.  Just use some comments to track what code is of what license,
like any intelligent and rational person might do.   After all, no sane person
thinks that if I glue two books together, I've made a derivative work and
require licensing to sell the result.  Were that the case, it would be illegal
to sell used books with notes written in the margins or papers inserted between
the pages.  Let's not be silly about this.  We all know what it means to derive
something from something else. 

...or, in simpler terms, this license is specifically incompatible with the GNU
GPL, but only because the GNU GPL requires that the GNU GPL be applied to all
source code which is part of a program that includes GNU GPL licensed source
code.  Many other free software licenses, such as the BSD license, make no such
demands and are therefore compatible with this license.   Only licenses which
require the relicensing of this source code are incompatible with this license.
Such licenses are viral in nature, and thus, this is the anti-viral license.*/


#include "global.h"

static time_t last_announce = 0;
void announce_server (int force) {
    struct sockaddr_in target;
    struct server_xmit_data announce;
    struct server_acknowledge_packet acknowledge;
    #ifdef LINUX    
    int ret = 0;
    #endif
    int a;
    time_t t;
	
    if (config.announce_disabled) return;


    if (force == 1) last_announce = 0;
    
   time (&t);
    
    if (t - last_announce < ANNOUNCE_SEND_INTERVAL) return;
    last_announce = t;
    
	if (!config.online) {	
		log_printf (NULL, LWARNING, "Server is offline");
		return;
	}
    
    _strncpy (announce.name, config.name, 64);
    announce.connections_max = config.connection_limit;
    announce.connections_current = 0;
    announce.server_type = config.server_type;
	_strncpy (announce.address, config.external_address, 256);
	memmove (&announce.salt, &config.salt, 20);

    #ifdef LINUX
/*
	if (g_announce_child) {
		printf ("announce already in progress\n");
		return;
	}*/
    g_announce_child ++;
    int pid = fork();
    if (pid != 0) {return;}
    if (g_bind_socket != -1) close (g_bind_socket);

    #endif
    
    if (lua_hook_call ("mme_global_event_announce", NULL)) return;
	
   	for (a = 1; a < NUM_USERS; a++) {
       	if (user[a].current_mode != USER_NOT_CONNECTED) close (user[a].socket);
		if (user[a].current_mode == USER_CONNECTED) announce.connections_current++;
   	}
    
    int port = 44436;
    char addr[1024];
    strcpy (addr, config.announce_server);
    char *cp;
    cp = strchr (addr, ':');
    if (cp != NULL) {
        *cp = 0;
        port = atoi (&cp[1]);
    }
    
    int sock = socket_connect (addr, port, &target);
    
    if (sock == -1) {
        log_printf (NULL, LWARNING, "Unable to connect to the server list: %s", strerror (errno));
        goto end;
    }
    
    announce.size = sizeof (announce);
	
	
	socket_ip_local_get (sock, announce.local_address, 64);
	

    if (socket_send_all (sock, &announce, sizeof (struct server_xmit_data), 0) == -1) {
        log_printf (NULL, LWARNING, "Data not transmitted to server list");
        goto end;
    }
    
    
    if (socket_recv_all (sock, &acknowledge, sizeof (struct server_acknowledge_packet), 0) == -1) {
        log_printf (NULL, LWARNING, "Data not received from server list");
        goto end;
    }
    
	if (acknowledge.result != ANNOUNCE_OK) {
        char err_result[64];
        
        switch (acknowledge.result) {
        case ANNOUNCE_NAME_TOO_SHORT: _strncpy (err_result, "ANNOUNCE_NAME_TOO_SHORT", 64); break;
        case ANNOUNCE_BAD_ADDRESS: _strncpy (err_result, "ANNOUNCE_BAD_ADDRESS", 64); break;
        case ANNOUNCE_BAD_PORT: _strncpy (err_result, "ANNOUNCE_BAD_PORT", 64); break;
        case ANNOUNCE_I_HATE_YOU: _strncpy (err_result, "ANNOUNCE_I_HATE_YOU", 64); break;
        case ANNOUNCE_TOO_EARLY: _strncpy (err_result, "ANNOUNCE_TOO_EARLY", 64); break;
        case ANNOUNCE_BAD_MATH: _strncpy (err_result, "ANNOUNCE_BAD_MATH", 64); break;
        case ANNOUNCE_INVALID_LENGTH: _strncpy (err_result, "ANNOUNCE_INVALID_LENGTH", 64); break;
        case ANNOUNCE_LIST_FULL: _strncpy (err_result, "ANNOUNCE_LIST_FULL", 64); break;
        default:
            snprintf (err_result, 63, "Unknown %i", acknowledge.result);
        }
    
        log_printf (NULL, LWARNING, "Announcement failed; result was %i: %s", acknowledge.result, err_result);
    
    }
    
    #ifdef LINUX
    ret = 1;
    #endif
    
    end:
    if (sock != -1) closesocket (sock);

#ifdef LINUX

    exit (ret);
#endif
    return;
}
